import React, { Component } from 'react'
import validate from 'validate.js'
import PropTypes from 'prop-types'

class Field extends Component {
	constructor(props) {
		super(props)

		this.state = {
			hasBlurred: false,
			attribs: {
				// Apply bindings, passing in existing events, so they also get called.
				onChange: this.handleOnChange.bind(this, props.onChange, props.getComponentValue),
				onBlur: this.handleOnBlur.bind(this, props.onBlur)
			}
		}

		this.debug = this.debug.bind(this)
		this.getValue = this.getValue.bind(this)
		this.isDebug = this.isDebug.bind(this)
		this.validate = this.validate.bind(this)
		this.isMountedComponent = false
	}

	componentDidMount() {
		this.context.store.setFieldRef(this)
		this.isMountedComponent = true

		this.debug({ fieldCreated: {
			props: this.props
		} })
	}

	componentWillUnmount() {
		this.context.store.deleteFieldRef(this)
		this.isMountedComponent = false

		this.debug({ fieldRemoved: {
			props: this.props
		} })
	}

	getValue(constraints) {
		let value = this.context.store && this.context.store.get && this.context.store.get(this.props.name, undefined, !!(constraints && constraints.allowEmpty))
		if (this.props.arrayIndex !== undefined) {
			value = value[this.props.arrayIndex]
		}
		return value
	}

	//	For numbers we need value to be passed back as undefined, and it works with other validations as well.
	//	If you want to always get an undefined value (Note: numeric values break on empty!), add allowEmpty in your validation
	getValueForValidation(constraints) {
		let value = this.getValue()
		if (!(constraints && constraints.allowEmpty) && value === '') {
			value = undefined
		}
		return value
	}

	isDebug() {
		return (this.props.debug === undefined && this.context.debug) || this.props.debug
	}

	debug(obj) {
		if (this.isDebug()) {
			console.log(`Formality field ${this.props.name}${this.props.arrayIndex !== undefined ? `[${this.props.arrayIndex}]` : ''}`, obj) // eslint-disable-line no-console
		}
	}

	async validate(setValidation = true) {
		//	Validates the field
		const obj = {}
		const constraints = {}
		const name = this.props.name
		const value = this.getValueForValidation(this.props.validate)

		obj[name] = value
		constraints[name] = this.props.validate
		let validationErrors
		try {
			await validate.async(obj, constraints)
		} catch (err) {
			validationErrors = err instanceof Error ? ['Unknown error'] : err[name]

			this.debug({ validateFieldError: {
				exception: err
			} })
		}

		// during form submit, we can let the form set all the validations at the end
		if (setValidation) {
			this.context.store.setValidation(name, validationErrors)
		}

		const error = validationErrors && validationErrors.length > 0
		const helperText = error ? validationErrors.join(', ') : null
		const errorAttribs = this.props.supportsHelperText ? { error, helperText } : { error }
		if (this.isMountedComponent) {
			this.setState({ attribs: { ...this.state.attribs, ...errorAttribs } })
		}

		this.debug({ validateField: {
			value,
			validate: this.props.validate,
			isValid: !validationErrors,
			validationResult: validationErrors,
			...errorAttribs
		} })

		return validationErrors // returns the validation result so the Form can call this validate function
	}

	async handleOnChange(oldHandler, getComponentValue, ...args) {
		const [event, index, setValue] = args
		//	When the field's onChange event fires, we update the state value
		const name = this.props.name
		let value

		//	Ability to use custom handler for getting the value
		if (getComponentValue) {
			value = getComponentValue.apply(this, args)
		} else {
			//	Get the value from the event - or the `setValue`, or `index`, if no
			//	setValue. This caters for components that don't always return the
			//	value in the event, such as material-ui. The `index` and `setValue`
			//	is for indexed components, such a a dropdown menu, but of course
			//	works with standard components that return the value in the `index`
			//	variable.
			// eslint-disable-next-line no-nested-ternary
			value = (event && event.target && event.target.value !== undefined) ?
				event.target.value : setValue !== undefined ?
				setValue : index
		}

		this.context.store.set(name, value, this.props.arrayIndex)

		//	Validate, as long as the field has been blurred at least once
		if (this.props.validate && (this.props.forcevalidate || this.state.hasBlurred)) {
			await this.validate()
		}

		if (oldHandler) {
			oldHandler.apply(this, args)
		}

		this.debug({ handleOnChange: {
			value,
			store: this.context.store,
			validated: this.props.validate && this.state.hasBlurred,
			hasOldHandler: !!oldHandler
		} })
	}

	async handleOnBlur(oldHandler, ...args) {
		//	When the field is blurred - we use this to see if we should validate
		//	TODO: need a way to do this on non-blurrable nodes, such as the RadioButtonGroup example...
		if (this.isMountedComponent) {
			this.setState({ hasBlurred: true })
		}
		if (this.props.validate) {
			await this.validate()
		}

		if (oldHandler) {
			oldHandler.apply(this, args)
		}

		this.debug({ handleOnBlur: {
			state: this.state,
			validated: this.props.validate,
			hasOldHandler: !!oldHandler
		} })
	}

	render() {
		if (!this.context.store) {
			throw new Error('You must have an enclosing form with a store to use this field component')
		}

		//	Grab the props - override with the state, so we can set the helperText, etc...
		const attribs = { ...this.props, ...this.state.attribs }

		if (!this.props.skipAutosetValue && this.props.value === undefined) {
			attribs.value = this.getValue(this.props.validate)
		}

		//	Remove things React doesn't know about
		delete attribs.component
		delete attribs.arrayIndex
		delete attribs.store
		delete attribs.getComponentValue
		delete attribs.validate
		delete attribs.forcevalidate
		delete attribs.skipAutosetValue
		delete attribs.debug
		delete attribs.supportsHelperText

		const myComponent = React.createElement(this.props.component, attribs, this.props.children)

		return myComponent
	}
}

Field.contextTypes = {
	store: PropTypes.object,
	debug: PropTypes.bool
}

Field.propTypes = {
	supportsHelperText: PropTypes.bool,
	debug: PropTypes.bool,
	name: PropTypes.string.isRequired,
	arrayIndex: PropTypes.number,
	validate: PropTypes.object, // eslint-disable-line react/forbid-prop-types
	onChange: PropTypes.func,
	getComponentValue: PropTypes.func,
	onBlur: PropTypes.func,
	component: PropTypes.any.isRequired, // eslint-disable-line react/forbid-prop-types
	value: PropTypes.any, // eslint-disable-line react/forbid-prop-types, the type is up to the wrapped component
	skipAutosetValue: PropTypes.bool,
	children: PropTypes.any, // eslint-disable-line react/forbid-prop-types
}

Field.defaultProps = {
	supportsHelperText: true,
	debug: undefined,
	arrayIndex: undefined,
	validate: undefined,
	onChange: undefined,
	getComponentValue: undefined,
	onBlur: undefined,
	value: undefined,
	skipAutosetValue: false,
	children: undefined
}

export default Field
